from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    import sys

    if sys.version_info <= (3, 10):
        from typing_extensions import Required
    else:
        from typing import Required


class _BaseURLDict(t.TypedDict):
    """Base direct URL data."""

    url: str


class VCSInfoDict(t.TypedDict, total=False):
    """VCS information."""

    vcs: Required[str]
    commit_id: Required[str]
    requested_revision: str
    resolved_revision: str
    resolved_revision_type: str


class VCSDict(t.TypedDict):
    """VCS direct URL data."""

    url: str
    vcs_info: VCSInfoDict


class ArchiveInfoDict(t.TypedDict, total=False):
    """Archive information."""

    hashes: dict[str, str]
    hash: str


class ArchiveDict(_BaseURLDict):
    """Archive direct URL data."""

    archive_info: ArchiveInfoDict


class DirectoryInfoDict(t.TypedDict, total=False):
    """Local directory information."""

    editable: bool


class DirectoryDict(_BaseURLDict):
    """Local directory direct URL data."""

    dir_info: DirectoryInfoDict



"""Sphinx configuration."""

from __future__ import annotations

import pep610

# Add any Sphinx extension module names here, as strings.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "myst_parser",
    "sphinx_design",
]

# General information about the project.
project = "pep610"
author = "Edgar Ramírez-Mondragón"
version = pep610.__version__
release = pep610.__version__
project_copyright = f"2023, {author}"

# -- Options for HTML output --------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "furo"
html_title = "PEP 610 - Direct URL Parser and Builder for Python"
html_theme_options = {
    "navigation_with_keys": True,
    "source_repository": "https://github.com/edgarrmondragon/citric/",
    "source_branch": "main",
    "source_directory": "docs/",
}

# -- Options for autodoc ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

autodoc_member_order = "bysource"
autodoc_preserve_defaults = True

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

# Only document types for parameters or return values that are already documented by the
# docstring.
autodoc_typehints_description_target = "documented"

# -- Options for extlinks -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html

extlinks_detect_hardcoded_links = True
extlinks = {
    "spec": (
        "https://packaging.python.org/en/latest/specifications/direct-url-data-structure/#%s-urls",
        "specification for %s URLs",
    ),
}

# -- Options for intersphinx ----------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration
intersphinx_mapping = {
    "metadata": ("https://importlib-metadata.readthedocs.io/en/latest", None),
    "python": ("https://docs.python.org/3/", None),
}

# -- Options for Myst Parser -------------------------------------------------------
# https://myst-parser.readthedocs.io/en/latest/configuration.html
myst_enable_extensions = ["colon_fence"]



import json
from importlib.metadata import Distribution

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema

from pep610 import SCHEMA_FILE, read_from_distribution, write_to_distribution

SCHEMA = json.loads(SCHEMA_FILE.read_text())


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(from_schema(SCHEMA))
def test_generic(tmp_path_factory: pytest.TempPathFactory, value: dict):
    """Test parsing a local directory."""
    dist_path = tmp_path_factory.mktemp("pep610")
    dist = Distribution.at(dist_path)
    write_to_distribution(dist, value)
    assert read_from_distribution(dist) is not None



"""PEP 610 parser."""

from __future__ import annotations

import json
import sys
import typing as t
from dataclasses import dataclass
from functools import singledispatch
from importlib.metadata import version

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources

if t.TYPE_CHECKING:
    from importlib.metadata import Distribution, PathDistribution

    if sys.version_info <= (3, 10):
        from typing_extensions import Self
    else:
        from typing import Self

    from pep610._types import (
        ArchiveDict,
        ArchiveInfoDict,
        DirectoryDict,
        DirectoryInfoDict,
        VCSDict,
        VCSInfoDict,
    )

__all__ = [
    "ArchiveData",
    "ArchiveInfo",
    "DirData",
    "DirInfo",
    "HashData",
    "VCSData",
    "VCSInfo",
    "__version__",
    "read_from_distribution",
    "to_dict",
    "write_to_distribution",
]

SCHEMA_FILE = importlib_resources.files(__package__) / "direct_url.schema.json"
__version__ = version(__package__)


@dataclass
class VCSInfo:
    """VCS information.

    See also :spec:`vcs`.

    Args:
        vcs: The VCS type.
        commit_id: The exact commit/revision number that was/is to be installed.
        requested_revision: A branch/tag/ref/commit/revision/etc (in a format
            compatible with the VCS).
    """

    vcs: str
    commit_id: str
    requested_revision: str | None = None
    resolved_revision: str | None = None
    resolved_revision_type: str | None = None


@dataclass
class _BaseData:
    """Base direct URL data.

    Args:
        url: The direct URL.
    """

    url: str


@dataclass
class VCSData(_BaseData):
    """VCS direct URL data.

    Args:
        url: The VCS URL.
        vcs_info: VCS information.
    """

    vcs_info: VCSInfo


class HashData(t.NamedTuple):
    """(Deprecated) Archive hash data.

    Args:
        algorithm: The hash algorithm.
        value: The hash value.
    """

    algorithm: str
    value: str


@dataclass
class ArchiveInfo:
    """Archive information.

    See also :spec:`archive`.

    Args:
        hashes: Dictionary mapping a hash name to a hex encoded digest of the file.
        hash: The archive hash (deprecated).
    """

    hashes: dict[str, str] | None = None
    hash: HashData | None = None


@dataclass
class ArchiveData(_BaseData):
    """Archive direct URL data.

    Args:
        url: The archive URL.
        archive_info: Archive information.
    """

    archive_info: ArchiveInfo


@dataclass
class DirInfo:
    """Local directory information.

    See also :spec:`directory`.

    Args:
        editable: Whether the distribution is installed in editable mode.
    """

    editable: bool | None

    def is_editable(self: Self) -> bool:
        """Distribution is editable?

        ``True`` if the distribution was/is to be installed in editable mode,
        ``False`` otherwise. If absent, default to ``False``

        Returns:
            Whether the distribution is installed in editable mode.
        """
        return self.editable is True


@dataclass
class DirData(_BaseData):
    """Local directory direct URL data.

    Args:
        url: The local directory URL.
        dir_info: Local directory information.
    """

    dir_info: DirInfo


@singledispatch
def to_dict(data) -> dict[str, t.Any]:  # noqa: ANN001
    """Convert the parsed data to a dictionary.

    Args:
        data: The parsed data.

    Raises:
        NotImplementedError: If the data type is not supported.
    """
    message = f"Cannot serialize unknown direct URL data of type {type(data)}"
    raise NotImplementedError(message)


@to_dict.register(VCSData)
def _(data: VCSData) -> VCSDict:
    vcs_info: VCSInfoDict = {
        "vcs": data.vcs_info.vcs,
        "commit_id": data.vcs_info.commit_id,
    }
    if data.vcs_info.requested_revision is not None:
        vcs_info["requested_revision"] = data.vcs_info.requested_revision
    if data.vcs_info.resolved_revision is not None:
        vcs_info["resolved_revision"] = data.vcs_info.resolved_revision
    if data.vcs_info.resolved_revision_type is not None:
        vcs_info["resolved_revision_type"] = data.vcs_info.resolved_revision_type

    return {"url": data.url, "vcs_info": vcs_info}


@to_dict.register(ArchiveData)
def _(data: ArchiveData) -> ArchiveDict:
    archive_info: ArchiveInfoDict = {}
    if data.archive_info.hashes is not None:
        archive_info["hashes"] = data.archive_info.hashes

    if data.archive_info.hash is not None:
        archive_info["hash"] = f"{data.archive_info.hash.algorithm}={data.archive_info.hash.value}"

    return {"url": data.url, "archive_info": archive_info}


@to_dict.register(DirData)
def _(data: DirData) -> DirectoryDict:
    dir_info: DirectoryInfoDict = {}
    if data.dir_info.editable is not None:
        dir_info["editable"] = data.dir_info.editable
    return {"url": data.url, "dir_info": dir_info}


def _parse(content: str) -> VCSData | ArchiveData | DirData | None:
    data = json.loads(content)

    if "archive_info" in data:
        hashes = data["archive_info"].get("hashes")
        hash_data = None
        if hash_value := data["archive_info"].get("hash"):
            hash_data = HashData(*hash_value.split("=", 1)) if hash_value else None

        return ArchiveData(
            url=data["url"],
            archive_info=ArchiveInfo(hashes=hashes, hash=hash_data),
        )

    if "dir_info" in data:
        return DirData(
            url=data["url"],
            dir_info=DirInfo(
                editable=data["dir_info"].get("editable"),
            ),
        )

    if "vcs_info" in data:
        return VCSData(
            url=data["url"],
            vcs_info=VCSInfo(
                vcs=data["vcs_info"]["vcs"],
                commit_id=data["vcs_info"]["commit_id"],
                requested_revision=data["vcs_info"].get("requested_revision"),
                resolved_revision=data["vcs_info"].get("resolved_revision"),
                resolved_revision_type=data["vcs_info"].get("resolved_revision_type"),
            ),
        )

    return None


def read_from_distribution(dist: Distribution) -> VCSData | ArchiveData | DirData | None:
    """Read the package data for a given package.

    Args:
        dist(importlib_metadata.Distribution): The package distribution.

    Returns:
        The parsed PEP 610 file.
    """
    if contents := dist.read_text("direct_url.json"):
        return _parse(contents)

    return None


def write_to_distribution(dist: PathDistribution, data: dict) -> int:
    """Write the direct URL data to a distribution.

    Args:
        dist: The distribution.
        data: The direct URL data.

    Returns:
        The number of bytes written.
    """
    return dist._path.joinpath(  # type: ignore[attr-defined]  # noqa: SLF001
        "direct_url.json",
    ).write_text(json.dumps(data))



"""Test the PEP 610 parser."""

from __future__ import annotations

import typing as t
from importlib.metadata import Distribution

import pytest

from pep610 import (
    ArchiveData,
    ArchiveInfo,
    DirData,
    DirInfo,
    HashData,
    VCSData,
    VCSInfo,
    read_from_distribution,
    to_dict,
    write_to_distribution,
)

if t.TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        pytest.param(
            {"url": "file:///home/user/project", "dir_info": {"editable": True}},
            DirData(
                url="file:///home/user/project",
                dir_info=DirInfo(editable=True),
            ),
            id="local_editable",
        ),
        pytest.param(
            {"url": "file:///home/user/project", "dir_info": {"editable": False}},
            DirData(
                url="file:///home/user/project",
                dir_info=DirInfo(editable=False),
            ),
            id="local_not_editable",
        ),
        pytest.param(
            {"url": "file:///home/user/project", "dir_info": {}},
            DirData(
                url="file:///home/user/project",
                dir_info=DirInfo(editable=None),
            ),
            id="local_no_editable_info",
        ),
        pytest.param(
            {
                "url": "https://github.com/pypa/pip/archive/1.3.1.zip",
                "archive_info": {
                    "hashes": {
                        "md5": "c4e0f0a1e0a5e708c8e3e3c4cbe2e85f",
                        "sha256": "2dc6b5a470a1bde68946f263f1af1515a2574a150a30d6ce02c6ff742fcc0db8",  # noqa: E501
                    }
                },
            },
            ArchiveData(
                url="https://github.com/pypa/pip/archive/1.3.1.zip",
                archive_info=ArchiveInfo(
                    hashes={
                        "md5": "c4e0f0a1e0a5e708c8e3e3c4cbe2e85f",
                        "sha256": "2dc6b5a470a1bde68946f263f1af1515a2574a150a30d6ce02c6ff742fcc0db8",  # noqa: E501
                    },
                ),
            ),
            id="archive_hashes",
        ),
        pytest.param(
            {
                "url": "https://github.com/pypa/pip/archive/1.3.1.zip",
                "archive_info": {
                    "hash": "sha256=2dc6b5a470a1bde68946f263f1af1515a2574a150a30d6ce02c6ff742fcc0db8",  # noqa: E501
                },
            },
            ArchiveData(
                url="https://github.com/pypa/pip/archive/1.3.1.zip",
                archive_info=ArchiveInfo(
                    hash=HashData(
                        "sha256",
                        "2dc6b5a470a1bde68946f263f1af1515a2574a150a30d6ce02c6ff742fcc0db8",
                    ),
                ),
            ),
            id="archive_sha256_legacy",
        ),
        pytest.param(
            {
                "url": "file://path/to/my.whl",
                "archive_info": {},
            },
            ArchiveData(
                url="file://path/to/my.whl",
                archive_info=ArchiveInfo(hash=None),
            ),
            id="archive_no_hashes",
        ),
        pytest.param(
            {
                "url": "https://github.com/pypa/pip.git",
                "vcs_info": {
                    "vcs": "git",
                    "requested_revision": "1.3.1",
                    "resolved_revision_type": "tag",
                    "commit_id": "7921be1537eac1e97bc40179a57f0349c2aee67d",
                },
            },
            VCSData(
                url="https://github.com/pypa/pip.git",
                vcs_info=VCSInfo(
                    vcs="git",
                    requested_revision="1.3.1",
                    resolved_revision_type="tag",
                    commit_id="7921be1537eac1e97bc40179a57f0349c2aee67d",
                ),
            ),
            id="vcs_git",
        ),
        pytest.param(
            {
                "url": "https://github.com/pypa/pip.git",
                "vcs_info": {
                    "vcs": "git",
                    "resolved_revision_type": "tag",
                    "commit_id": "7921be1537eac1e97bc40179a57f0349c2aee67d",
                },
            },
            VCSData(
                url="https://github.com/pypa/pip.git",
                vcs_info=VCSInfo(
                    vcs="git",
                    requested_revision=None,
                    resolved_revision_type="tag",
                    commit_id="7921be1537eac1e97bc40179a57f0349c2aee67d",
                ),
            ),
            id="vcs_git_no_requested_revision",
        ),
        pytest.param(
            {
                "url": "https://github.com/pypa/pip.git",
                "vcs_info": {
                    "vcs": "git",
                    "requested_revision": "1.3.1",
                    "resolved_revision": "1.3.1",
                    "resolved_revision_type": "tag",
                    "commit_id": "7921be1537eac1e97bc40179a57f0349c2aee67d",
                },
            },
            VCSData(
                url="https://github.com/pypa/pip.git",
                vcs_info=VCSInfo(
                    vcs="git",
                    requested_revision="1.3.1",
                    resolved_revision="1.3.1",
                    resolved_revision_type="tag",
                    commit_id="7921be1537eac1e97bc40179a57f0349c2aee67d",
                ),
            ),
            id="vcs_git_resolved_revision",
        ),
        pytest.param(
            {
                "url": "https://github.com/pypa/pip.git",
                "vcs_info": {
                    "vcs": "git",
                    "requested_revision": "1.3.1",
                    "resolved_revision": "1.3.1",
                    "commit_id": "7921be1537eac1e97bc40179a57f0349c2aee67d",
                },
            },
            VCSData(
                url="https://github.com/pypa/pip.git",
                vcs_info=VCSInfo(
                    vcs="git",
                    requested_revision="1.3.1",
                    resolved_revision="1.3.1",
                    resolved_revision_type=None,
                    commit_id="7921be1537eac1e97bc40179a57f0349c2aee67d",
                ),
            ),
            id="vcs_no_resolved_revision",
        ),
    ],
)
def test_parse(data: dict, expected: object, tmp_path: Path):
    """Test the parse function."""
    dist = Distribution.at(tmp_path)
    write_to_distribution(dist, data)

    result = read_from_distribution(dist)
    assert result == expected

    assert to_dict(result) == data


def test_unknown_data_type():
    """Test serialization from unknown data fails."""
    data = object()
    with pytest.raises(NotImplementedError, match="Cannot serialize unknown"):
        to_dict(data)


def test_local_directory(tmp_path: Path):
    """Test that a local directory is read back as a local directory."""
    data = {
        "url": "file:///home/user/project",
        "dir_info": {"editable": True},
    }
    dist = Distribution.at(tmp_path)
    write_to_distribution(dist, data)

    result = read_from_distribution(dist)
    assert isinstance(result, DirData)
    assert result.url == "file:///home/user/project"
    assert result.dir_info.is_editable()
    assert to_dict(result) == data

    result.dir_info.editable = False
    assert to_dict(result) == {
        "url": "file:///home/user/project",
        "dir_info": {"editable": False},
    }

    result.dir_info.editable = None
    assert to_dict(result) == {
        "url": "file:///home/user/project",
        "dir_info": {},
    }


def test_unknown_url_type(tmp_path: Path):
    """Test that an unknown URL type is read back as None."""
    data = {
        "url": "unknown:///home/user/project",
        "unknown_info": {},
    }
    dist = Distribution.at(tmp_path)
    write_to_distribution(dist, data)
    assert read_from_distribution(dist) is None


def test_no_file(tmp_path: Path):
    """Test that a missing file is read back as None."""
    dist = Distribution.at(tmp_path)
    assert read_from_distribution(dist) is None
